import json
import logging
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import AuthException, ConfigException
import socket
import time
import re
import requests
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter

logger = logging.getLogger(__name__)
MAX_RESULTS = 1000
MAX_EVENTS = 99

class CustomISAMPluginRemote(RemoteBasePlugin):
    
    previousEvents = []
    events = 0
    
    def query(self, **kwargs):
        logger.info(f'Polling to ISAM API')
        config = kwargs['config']
        
        logging.basicConfig(level=logging.DEBUG)
        logger.setLevel(logging.DEBUG)

            
        # Config GUI vars
        host_port = config['host_port'].strip()
        host_type = config['host_type']
        self.user = config['user'].strip()
        self.password = config['password']
        self.second_user = config['second_user'].strip()
        self.second_password = config['second_password']
        high_proxies = config['high_reverse_proxies'].strip()
        high_junctions = config['high_junctions'].strip()
        get_threads = True if 'get_threads' in config and str(config['get_threads']).lower() == 'true' else False
        
        group = config['group']
        
        start_time = int(time.time())-60
        
        g1 = self.topology_builder.create_group(f'ISAM - {group}', f'ISAM - {group}')
        e1 = g1.create_element(f'ISAM - {group} - {host_port}', f'ISAM - {group} - {host_port}')
        e1.report_property('Plugin version', kwargs['json_config']['version'])
        
        metrics = []
        metrics.append({'name': 'total_memory', 'owner': 'host', 'category': 'memory', 'metric_type': 'absolute', 'field': 'total'})
        metrics.append({'name': 'free_memory', 'owner': 'host', 'category': 'memory', 'metric_type': 'absolute', 'field': 'free'})
        metrics.append({'name': 'used_memory', 'owner': 'host', 'category': 'memory', 'metric_type': 'absolute', 'field': 'used'})
        metrics.append({'name': 'free_memory_pct', 'owner': 'host', 'category': 'memory', 'metric_type': 'absolute', 'calculate': '([free]/[total])*100'})
        metrics.append({'name': 'used_memory_pct', 'owner': 'host', 'category': 'memory', 'metric_type': 'absolute', 'calculate': '(([total]-[free])/[total])*100'})
        metrics.append({'name': 'user_cpu', 'owner': 'host', 'category': 'cpu', 'metric_type': 'absolute', 'field': 'user_cpu'})
        metrics.append({'name': 'system_cpu', 'owner': 'host', 'category': 'cpu', 'metric_type': 'absolute', 'field': 'system_cpu'})
        metrics.append({'name': 'idle_cpu', 'owner': 'host', 'category': 'cpu', 'metric_type': 'absolute', 'field': 'idle_cpu'})
        metrics.append({'name': 'root_total_size', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'size', 'location': 'root'})
        metrics.append({'name': 'root_used_space', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'used', 'location': 'root'})
        metrics.append({'name': 'root_free_space', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'avail', 'location': 'root'})
        metrics.append({'name': 'boot_total_size', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'size', 'location': 'boot'})
        metrics.append({'name': 'boot_used_space', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'used', 'location': 'boot'})
        metrics.append({'name': 'boot_free_space', 'owner': 'host', 'category': 'storage', 'metric_type': 'absolute', 'field': 'avail', 'location': 'boot'})
        metrics.append({'name': 'proxy_threads_total', 'owner': 'proxy', 'category': 'thread', 'metric_type': 'absolute', 'field': 'total', 'split_name': 'Instance', 'method': 'POST'})
        metrics.append({'name': 'proxy_threads_active', 'owner': 'proxy', 'category': 'thread', 'metric_type': 'absolute', 'field': 'active', 'split_name': 'Instance', 'method': 'POST'})
        metrics.append({'name': 'junction_active_worker_threads', 'owner': 'junction', 'category': 'thread', 'metric_type': 'absolute', 'field': 'Active worker threads', 'split_name': 'Junction', 'method': 'POST'})
        metrics.append({'name': 'proxy_requests', 'owner': 'proxy', 'category': 'throughput', 'metric_type': 'absolute', 'field': 'total_hits', 'split_by': 'instance', 'split_name': 'Instance'})
        metrics.append({'name': 'proxy_throughput', 'owner': 'proxy', 'category': 'throughput', 'metric_type': 'absolute', 'calculate': '[total_hits]/60', 'split_by': 'instance', 'split_name': 'Instance'})
        metrics.append({'name': 'junction_response_time', 'owner': 'junction', 'category': 'junction_response', 'metric_type': 'absolute', 'field': 'r', 'split_by': 'junction', 'split_name': 'Junction'})
        metrics.append({'name': 'junction_requests', 'owner': 'junction', 'category': 'junction_requests', 'metric_type': 'absolute', 'field': 'h', 'split_by': 'a', 'split_name': 'Junction'})
        
        categories = {
                        'memory': {'url': f'https://{host_port}/statistics/systems/memory.json?timespan=1h'},
                        'cpu': {'url': f'https://{host_port}/statistics/systems/cpu.json?timespan=1h'},
                        'storage': {'url': f'https://{host_port}/statistics/systems/storage.json?timespan=1h'},
                        'thread': {'url': f'https://{host_port}/isam/pdadmin', 'method': 'POST'},
                        'throughput': {'url': f'https://{host_port}/analysis/reverse_proxy_traffic/throughput_widget/?date={start_time}&duration=60'},
                        'junction_response': {'url': f'https://{host_port}/analysis/reverse_proxy_traffic/reqtime?duration=60&date={start_time}&instance=[instance]', 'resolve': 'instance'},
                        'junction_requests': {'url': f'https://{host_port}/analysis/reverse_proxy_traffic/traffic/instance/[instance]/?date={start_time}&duration=60&aspect=junction', 'resolve': 'instance'}
                     }
                     
        # Start getting proxies
        instances = []
        if high_proxies == '': # Get all
            list_url = f'https://{host_port}/wga/reverseproxy/'
            list_proxies_result = self.get_api_response(list_url)
            
            logger.info(f'List of proxies result: {list_proxies_result}')
            if list_proxies_result:
                for each in list_proxies_result:
                    if each['enabled'] == 'yes':
                        instances.append(each['instance_name'].strip())
        else:
            instances = high_proxies.split(',')
            
        # Get all junctions per instance
        junctions = {}
        if high_junctions == '': # Get all junctions per instance
            for instance in instances:
                list_url = f'https://{host_port}/wga/reverseproxy/{instance}/junctions'
                list_junctions_result = self.get_api_response(list_url)
                logger.info(f'List junctions result: {list_junctions_result}')
                if list_junctions_result:
                    for each in list_junctions_result:
                        if instance not in junctions:
                            junctions[instance] = []
                        junctions[instance].append(each['id'].strip())
        else:
            for instance in instances:
                junctions[instance] = []
                high_junctions_list = high_junctions.split(',')
                for each_junction in high_junctions_list:
                    junctions[instance].append(each_junction)
          
        if instances and junctions:
            # Build post_data
            post_data = {"admin_id": self.second_user,
                        "admin_pwd": self.second_password}
            
            commands = []
            for instance in instances:
                commands.append(f"server task {instance}-webseald-{host_port} stats get pdweb.threads")
                
            for instance in junctions:
                for junction in junctions[instance]:
                    commands.append(f"server task {instance}-webseald-{host_port} show {junction}")
                    
            post_data['commands'] = commands
            
            # Poll all categories and store
            category_responses = {}
            
            for each_cat in categories:
                response = None
                if 'method' in categories[each_cat] and categories[each_cat]['method'] == 'POST':
                    if get_threads:
                        response = self.post_api_response(categories[each_cat]['url'], post_data)
                        
                        logger.info(f'RAW POST response: {response}')
                        if response:
                            # For some reason the response is not json, it is a block of text that needs to be converted to json
                            response = self.parse_text_to_json(response)
                            category_responses[each_cat] = response
                else: # It's a GET
                    if '[instance]' in categories[each_cat]['url']:
                        for instance in instances:
                            category_per_instance_result = self.get_api_response(categories[each_cat]['url'].replace('[instance]', instance))
                            if category_per_instance_result:
                                if each_cat not in category_responses:
                                    category_responses[each_cat] = {}
                                category_responses[each_cat][instance] = category_per_instance_result
                    else:
                        response = self.get_api_response(categories[each_cat]['url'])
                        if response:
                            category_responses[each_cat] = response
                if response:
                    logger.info(f'{each_cat} response: {response}')
            
            for metric in metrics:
                if metric['category'] in category_responses:
                    logger.info(f"Category {metric['category']} response: {category_responses[metric['category']]}")
                    logger.info(f"metric: {metric}")
                    #if isinstance(category_responses[metric['category']], list):
                    #    category_responses[metric['category']] = category_responses[metric['category']][0]
                    if 'field' in metric and metric['field'] in category_responses[metric['category']]:
                        e1.absolute(key=metric['name'], value=float(category_responses[metric['category']][metric['field']]))
                    elif 'location' in metric:
                        e1.absolute(key=metric['name'], value=float(category_responses[metric['category']][metric['location']][metric['field']]))
                    elif 'method' in metric and metric['method'] == 'POST':
                        for each_cmd in category_responses[metric['category']]:
                            logger.info(f"each_cmd = {each_cmd}")
                            each_cmd_words = each_cmd.split()
                            instance = each_cmd_words[2][0:each_cmd_words[2].index('-')]
                            if metric['owner'] is 'proxy' and each_cmd.endswith('pdweb.threads'):
                                e1.absolute(key=metric['name'], value=float(category_responses[metric['category']][each_cmd][metric['field']]), dimensions={metric['split_name']: instance})
                            elif metric['owner'] is 'junction' and 'show /' in each_cmd:
                                e1.absolute(key=metric['name'], value=float(category_responses[metric['category']][each_cmd][metric['field']]), dimensions={'Instance': instance, 'Junction': category_responses[metric['category']][each_cmd]['Junction point']})
                            
                                # Send alert if server is not running
                                servers = [ v for k,v in category_responses[metric['category']][each_cmd].items() if k.startswith('Server ')]
                                logger.info(f"Servers: {servers}")
                                for each_server in servers:
                                    logger.info(f"each server: {each_server}")
                                    if each_server['Server State'] is not 'running':
                                        e1.report_availability_event(f'Server {each_server["Hostname"]} on port {each_server["Port"]} for junction {category_responses[metric["category"]][each_cmd]["Junction point"]} is not running', f'ISAM server for junction not running')
                    elif metric['category'] is 'junction_response':
                        for each_instance in category_responses[metric['category']]:
                            logger.info(f"each_instance = {category_responses[metric['category']][each_instance]}")
                            if 'records' in category_responses[metric['category']][each_instance]:
                                for each_junction in category_responses[metric['category']][each_instance]['records']:
                                    logger.info(f"each_junction = {each_junction}")
                                    if 'records' in each_junction:
                                        if isinstance(each_junction['records'], list):
                                            each_junction['records'] = each_junction['records'][0]
                                        logger.info(f"junction_response -> each_junction = {each_junction}")
                                        if each_junction['records'] and metric['field'] in each_junction['records']:
                                            record = each_junction['records'][metric['field']]
                                            if record is not None:
                                                e1.absolute(key=metric['name'], value=float(record), dimensions={'Instance': each_instance, metric['split_name']:each_junction['junction']})
                    elif metric['category'] is 'junction_requests':
                        for each_instance in category_responses[metric['category']]:
                            for each_junction in category_responses[metric['category']][each_instance]:
                                e1.absolute(key=metric['name'], value=float(each_junction[metric['field']]), dimensions={'Instance': each_instance, metric['split_name']:each_junction[metric['split_by']]})
                        
                    if 'calculate' in metric:
                        calculate = metric['calculate']
                        to_replace = re.findall('\[\w+\]', calculate)
                        if isinstance(category_responses[metric['category']], list):
                            for each_item in category_responses[metric['category']]:
                                for r in to_replace:
                                    field = r.strip('[""]')
                                    logger.info(f"each_item = {each_item}")
                                    if field in each_item:
                                        field = each_item[field]
                                        calculate = calculate.replace(r, str(field))
                                try:
                                    val = float(eval(calculate))
                                    if 'split_by' in metric:
                                        #logger.info(f"Setting metric {metric['name']} for dimension {metric['split_name']} {each_item[metric['split_by']]} = {val}")
                                        e1.absolute(key=metric['name'], value=val, dimensions={metric['split_name']: each_item[metric['split_by']]})
                                    else:
                                        #logger.info(f"Setting metric {metric['name']} = {val}")
                                        e1.absolute(key=metric['name'], value=val)
                                except Exception:
                                    logger.error(f"Could not calculate value for {metric['name']}. Calculation: {calculate}")
                                    pass
                        
                        else:
                            for r in to_replace:
                                field = r.strip('[""]')
                                if field in category_responses[metric['category']]:
                                    field = category_responses[metric['category']][field]
                                    calculate = calculate.replace(r, str(field))
                            try:
                                val = float(eval(calculate))
                                if 'split_by' in metric:
                                    #logger.info(f"Setting metric {metric['name']} for dimension {metric['split_name']} {category_responses[metric['category']][metric['split_by']]} = {val}")
                                    e1.absolute(key=metric['name'], value=val, dimensions={metric['split_name']: category_responses[metric['category']][metric['split_by']]})
                                else:
                                    #logger.info(f"Setting metric {metric['name']} = {val}")
                                    e1.absolute(key=metric['name'], value=val)
                            except Exception:
                                logger.error(f"Could not calculate value for {metric['name']}. Calculation: {calculate}")
                                pass
        else:
            logger.error('*** No instances or junctions were retrieved. No statistics to show ***')
            self.report_custom_info_event(e1, 'No instances or junctions were retrieved. No data to show. Make sure user has enough permissions and ISAM server is accessible', 'No data collected')
            
        # Check certificate expiration
        url = f'https://{host_port}/events/system/eventid/WGAWA0*45W'
        certificate_response = self.get_api_response(url)
        if certificate_response:
            for each in certificate_response:
                if each['description'].startswith('Certificate expires in'):
                    days_left = int(each['description'].split()[3])
                    if days_left <= 7:
                        e1.report_error_event(each['description'], f'Certificate expires in {days_left} days')
                    else:
                        self.report_custom_info_event(e1, each['description'], f'Certificate expires in {days_left} days')
        
        
        logger.info(f'Finished polling ISAM to {host_port}')

    def get_api_response(self, url):
        try:
            logger.info(f'GET url: {url}')
            basicAuth = HTTPBasicAuth(self.user, self.password)
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json', 'Cache-Control': 'no-cache'}
            if 'aspect=junction' in url:
                headers['Range'] = 'items=0-24'
            response = requests.get(url, auth=basicAuth, headers=headers, verify=False, timeout=15)
            response.raise_for_status()
            logger.info(f'Response code: {response.status_code}')
        except requests.exceptions.HTTPError as e:
            logger.exception(f'Failed polling: {url}. {e}')
            raise ConfigException(f'Error calling API: {e}')
        except requests.exceptions.RequestException as e:
            logger.exception(f'Failed polling: {url}. {e}')
            raise ConfigException(f'Error requesting data from ISAM server. {e}')
            return False # The request failed
        
        if response.status_code == 200:
            logger.info(f'Response from {url}: {response.json()}')
            if 'response' in response.json():
                return response.json()['response']
            return response.json()
        return False

    def post_api_response(self, url, data):
        try:
            scrubbed_data = data
            scrubbed_data['admin_pwd'] = '*******'
            logger.info(f'POST: {url} with post data: {scrubbed_data}')
            basicAuth = HTTPBasicAuth(self.user, self.password)
            response = requests.post(url, data=data, auth=basicAuth, headers={'Content-Type': 'application/json', 'Accept': 'application/json'}, verify=False, timeout=15)
            response.raise_for_status()
            logger.info(f'Response code: {response.status_code}')
        except requests.exceptions.HTTPError as e:
            logger.exception(f'Failed polling: {url}. {e}')
            raise ConfigException(f'Error calling API: {e}')
        except requests.exceptions.RequestException as e:
            logger.exception(f'Failed polling: {url}. {e}')
            return False # The request failed
        
        if response.status_code == 200:
            logger.info(f'Response from {url}: {response.json()}')
            return response.json()
        return False
        
    def report_custom_info_event(self, e1, message, title):
        if self.events < MAX_EVENTS:
            if title + message not in self.previousEvents:
                self.events += 1
                e1.report_custom_info_event(message, title)
                self.previousEvents.append(title + message)
                
    def parse_text_to_json(self, response):
        cmd_response = response['result'].split('cmd> ')
        regex = re.compile(r"(?:^\s*)(.*?)(?:\s*):(?:\s*)(.*?)(?:\s*$)")
        json_text = '{'
        for each_cmd in cmd_response:
            if each_cmd is not '':
                each_cmd_parts = each_cmd.split("\n",1)
                command = each_cmd_parts[0]
                resp = each_cmd_parts[1].split('\n')
                json_text = json_text+'"'+command+'": {'
                in_server = False
                for index, each_resp in enumerate(resp):
                    if each_resp is not '':
                        each_resp = each_resp.replace('\t', '')
                        if bool(re.search(r'Server \d', each_resp)):
                            if in_server:
                                json_text = json_text[0:-2]+'},\n'
                            key_value = regex.sub(r'"\1": {',each_resp)
                            in_server = True
                        else:
                            key_value = regex.sub(r'"\1":"\2",\n',each_resp)
                        json_text = json_text + key_value
                if in_server:
                    json_text = json_text[0:-2]+'},\n'
                json_text = json_text[0:-2]+'},'
        json_text = json_text[0:-1]+'}'
        
        return json.loads(json_text)